/**
 *  Exercise 1 - Part II
 *
 *  Templated non-numeric MatrixClass
 */

#include <vector>
#include <iostream>
#include <iomanip>
#include <cstdlib>

template<class T>
class MatrixClass
{
public:
    // forward declaration of nested classes
    class Row;
    class RowIterator;
    class ColIterator;
    class ConstRow;
    class ConstRowIterator;
    class ConstColIterator;

    // Implement Iterators
    class ConstColIterator
    {
        const MatrixClass<T> & A;
        unsigned int _r;
        unsigned int _c;
    public:
        ConstColIterator (const MatrixClass<T> & a, unsigned int r, unsigned int c) : A(a), _r(r), _c(c) {}
        ConstColIterator (const ColIterator & i) : A(i.A), _r(i._r), _c(i._c) {}
        const T& operator * () const
        {
            return A(_r,_c);
        }
        ConstColIterator & operator ++()
        {
            _c++;
            return *this;
        }
        bool operator == (const ConstColIterator & i) const
        {
            return i._r == _r && i._c == _c;
        }
        bool operator != (const ConstColIterator & i) const
        {
            return ! operator==(i);
        }
        bool operator == (const ColIterator & i) const
        {
            return i._r == _r && i._c == _c;
        }
        bool operator != (const ColIterator & i) const
        {
            return ! operator==(i);
        }
        unsigned int col() const { return _c; }
    };

    class ColIterator
    {
        MatrixClass<T> & A;
        unsigned int _r;
        unsigned int _c;
        friend class ConstColIterator;
    public:
        ColIterator (MatrixClass<T> & a, unsigned int r, unsigned int c) : A(a), _r(r), _c(c) {}
        T& operator * ()
        {
            return A(_r,_c);
        }
        const T operator * () const
        {
            return A(_r,_c);
        }
        ColIterator & operator ++()
        {
            _c++;
            return *this;
        }
        bool operator == (const ColIterator & i) const
        {
            return i._r == _r && i._c == _c;
        }
        bool operator != (const ColIterator & i) const
        {
            return ! operator==(i);
        }
        bool operator == (const ConstColIterator & i) const
        {
            return (i != *this);
        }
        bool operator != (const ConstColIterator & i) const
        {
            return (i == *this);
        }
        unsigned int col() const { return _c; }
    };

    class ConstRowIterator
    {
        const MatrixClass<T> & A;
        unsigned int _r;
        mutable ConstRow _constRow;
    public:
        ConstRowIterator (const MatrixClass<T> & a, unsigned int r) : A(a), _r(r), _constRow(A,_r) {}
        ConstRowIterator (const RowIterator & i) : A(i.A), _r(i._r), _constRow(A,_r) {}
        const ConstRow & operator * () const
        {
            _constRow.setRow(_r);
            return _constRow;
        }
        const ConstRow * operator -> () const
        {
            return & (*(*this));
        }
        ConstRowIterator & operator ++()
        {
            _r++;
            return *this;
        }
        bool operator == (const RowIterator & i) const
        {
            return i._r == _r;
        }
        bool operator != (const RowIterator & i) const
        {
            return ! operator==(i);
        }
        bool operator == (const ConstRowIterator & i) const
        {
            return i._r == _r;
        }
        bool operator != (const ConstRowIterator & i) const
        {
            return ! operator==(i);
        }
        unsigned int row() const { return _r; }
    };

    class RowIterator
    {
        MatrixClass<T> & A;
        unsigned int _r;
        mutable ConstRow _constRow;
        mutable Row _row;
        friend class ConstRowIterator;
    public:
        RowIterator (MatrixClass<T> & a, unsigned int r) : A(a), _r(r), _constRow(A,r), _row(A,r) {}
        Row & operator * ()
        {
            _row.setRow(_r);
            return _row;
        }
        const ConstRow & operator * () const
        {
            _constRow.setRow(_r);
            return _constRow;
        }
        Row * operator -> ()
        {
            return & (*(*this));
        }
        const ConstRow * operator -> () const
        {
            return & (*(*this));
        }
        RowIterator & operator ++()
        {
            _r++;
            return *this;
        }
        bool operator == (const RowIterator & i) const
        {
            return i._r == _r;
        }
        bool operator != (const RowIterator & i) const
        {
            return ! operator==(i);
        }
        bool operator == (const ConstRowIterator & i) const
        {
            return (i != *this);
        }
        bool operator != (const ConstRowIterator & i) const
        {
            return (i == *this);
        }
        unsigned int row() const { return _r; }
    };

    // represent a single matrix row
    class ConstRow
    {
        const MatrixClass<T> & A;
        int _r;
        friend class ConstRowIterator;
        void setRow(int i) { _r = i; }
    public:
        ConstRow (const MatrixClass<T> & a, int r)
            : A(a), _r(r) {}
        ConstRow (const Row & r)
            : A(r.A), _r(r._r) {}
        const T & diag() const
        {
            return A(_r,_r);
        }
        ConstColIterator begin() const
        {
            return ConstColIterator(A,_r,0);
        };
        ConstColIterator end() const
        {
            return ConstColIterator(A,_r,A.Cols());
        };
    };

    class Row
    {
        MatrixClass<T> & A;
        int _r;
        friend class RowIterator;
        friend class ConstRowIterator;
        void setRow(int i) { _r = i; }
    public:
        Row (MatrixClass<T> & a, int r)
            : A(a), _r(r) {}
        T & diag()
        {
            return A(_r,_r);
        }
        const T & diag() const
        {
            return A(_r,_r);
        }
        ColIterator begin()
        {
            return ColIterator(A,_r,0);
        };
        ColIterator end()
        {
            return ColIterator(A,_r,A.Cols());
        };
        ConstColIterator begin() const
        {
            return ConstColIterator(A,_r,A.Cols());
        };
        ConstColIterator end() const
        {
            return ConstColIterator(A,A.Cols(),A.Cols());
        };
    };

    // Iterator access
    RowIterator begin()
    {
        return RowIterator(*this,0);
    }
    ConstRowIterator begin() const
    {
        return ConstRowIterator(*this,0);
    }
    RowIterator end()
    {
        return RowIterator(*this,Rows());
    }
    ConstRowIterator end() const
    {
        return ConstRowIterator(*this,Rows());
    }
    
    // Set number of matrix rows and columns and 
    // initialize matrix elements with a given value
    void Resize(int numRows, int numCols, const T &value=T() );
    
    // Access matrix element at position (i,j)
    T& operator()(int i, int j);
    const T& operator()(int i, int j) const;
    
    // Output matrix content
    void Print() const;
    
    // Returns number of matrix raws
    int Rows() const 
    {
        return numRows_;
    }
    
    // Returns number of matrix columns
    int Cols() const
    {
        return numCols_;
    }

    // Constructors
    
    MatrixClass(int numRows=0, int numCols=0, const T &value=T() ) :
      a_(numRows), numRows_(numRows), numCols_(numCols)
    {
      if ( ((numCols==0) && (numRows!=0)) || ((numCols!=0) && (numRows==0)) )
      {
        numRows_=0;
        numCols_=0;
        a_.resize(numRows_);
      }
      else
      {
          for (int i=0;i<numRows_;++i)
              a_[i].resize(numCols_);
      }
      for (int i=0;i<numRows;++i)
      {
          for (int j=0;j<numCols;++j)
              a_[i][j]=value;
      }
    }

  protected:
    // matrix elements
    std::vector<std::vector<T> > a_;
    // number of rows
    int numRows_;
    //number of columns
    int numCols_;
};


// Set number of matrix rows and columns and 
// initialize matrix elements with a given value
template<class T>
void MatrixClass<T>::Resize(int numRows, int numCols, const T &value)
{
  a_.resize(numRows);
  for (unsigned int i=0;i<a_.size();++i)
  {
    a_[i].resize(numCols);
    for (unsigned int j=0;j<a_[i].size();++j)
        a_[i][j]=value;
  }
  numRows_=numRows;
  numCols_=numCols;
}

// Access matrix element at position (i,j)
template<class T>
T& MatrixClass<T>::operator()(int i, int j)
{
    if ((i<0)||(i>=numRows_))
    {
        std::cerr << "Illegal row index " << i;
        std::cerr << " valid range is (0:" << numRows_-1 << ")";
        std::cerr << std::endl;
        exit(EXIT_FAILURE);
    }
    if ((j<0)||(j>=numCols_))
    {
        std::cerr << "Illegal column index " << j;
        std::cerr << " valid range is (0:" << numCols_-1 << ")";
        std::cerr << std::endl;
        exit(EXIT_FAILURE);
    }
    return a_[i][j];
}

template<class T>
const T& MatrixClass<T>::operator()(int i, int j) const
{
  if ((i<0)||(i>=numRows_))
  {
    std::cerr << "Illegal row index " << i;
    std::cerr << " valid range is (0:" << numRows_-1 << ")";
    std::cerr << std::endl;
    exit(EXIT_FAILURE);
  }
  if ((j<0)||(j>=numCols_))
  {
    std::cerr << "Illegal column index " << j;
    std::cerr << " valid range is (0:" << numCols_-1 << ")";
    std::cerr << std::endl;
    exit(EXIT_FAILURE);
  }
  return a_[i][j];
}

template<class T>
void MatrixClass<T>::Print() const
{
    std::cout << "(" << numRows_ << "x";
    std::cout << numCols_ << ") matrix:" << std::endl;
    for (int i=0;i<numRows_;++i)
    {
        std::cout << std::setprecision(3);
        for (int j=0;j<numCols_;++j)
            std::cout << std::setw(5) << a_[i][j] << " ";
        std::cout << std::endl;
    }
    std::cout << std::endl;
}
